use rocksdb::{Direction, ErrorKind, IteratorMode, OptimisticTransactionDB, Options, DB};
use std::str::{from_utf8, Utf8Error};

const DB_PATH: &str = "/tmp/rdb";
const TAG_NAME_FROM_ID: &str = "name(id)";

struct User {
    id: u64,
    name: String,
}

impl User {
    fn new(id: u64, name: &str) -> Self {
        Self {
            id,
            name: name.to_string(),
        }
    }

    // `name(id)\0` followed by a big-endian encoded id.
    fn tag_name_of_id(id: u64) -> Vec<u8> {
        [
            TAG_NAME_FROM_ID.as_bytes(),
            [0 as u8].as_slice(),
            id.to_be_bytes().as_slice(),
        ]
        .concat()
    }

    // `name(id)\1`
    fn terminator_name_of_id() -> Vec<u8> {
        [TAG_NAME_FROM_ID.as_bytes(), [1 as u8].as_slice()].concat()
    }

    // Either a string or a UTf-8 encoding error.
    fn from_value_name(bytes: Vec<u8>) -> Result<String, Utf8Error> {
        Ok(from_utf8(bytes.as_slice())?.to_string())
    }

    // `name(id)\0` followed by this object's big-endian encoded id.
    fn key_name_of_id(&self) -> Vec<u8> {
        Self::tag_name_of_id(self.id)
    }

    // A byte array of the name encoded in UTF-8.
    fn value_name(&self) -> Vec<u8> {
        self.name.as_bytes().to_vec()
    }
}

fn main() {
    {
        // Create the database directory and files.
        let db: OptimisticTransactionDB = OptimisticTransactionDB::open_default(DB_PATH).unwrap();

        // Write user names.
        'write_txn: loop {
            let txn = db.transaction();

            // Note that names are not in id order.
            let user_name_list = vec![
                User::new(8, "ζ"),
                User::new(5, "ε"),
                User::new(21, "θ"),
                User::new(13, "η"),
                User::new(1, "β"),
                User::new(1, "α"),
                User::new(3, "δ"),
                User::new(2, "γ"),
            ];

            for user_name in user_name_list {
                txn.put(user_name.key_name_of_id(), user_name.value_name())
                    .unwrap();
            }

            // Commit the transaction.
            if let Err(e) = txn.commit() {
                if ErrorKind::Busy != e.kind() {
                    panic!("Can't commit transaction: {:?}", e);
                } else {
                    // TODO: exponential backoff.
                }
            } else {
                break 'write_txn;
            }
        }

        // Read user names.
        'read_txn: loop {
            let txn = db.transaction();

            let iter = txn.iterator(IteratorMode::From(
                &User::tag_name_of_id(0),
                Direction::Forward,
            ));
            let iter_end = User::terminator_name_of_id();

            'range_scan: for i in iter {
                let (key, value) = i.unwrap();
                if !(*key < *iter_end) {
                    break 'range_scan;
                }

                println!(
                    "key={:?} value={}",
                    key,
                    User::from_value_name(value.to_vec()).unwrap()
                );
            }

            // Commit the transaction.
            if let Err(e) = txn.commit() {
                if ErrorKind::Busy != e.kind() {
                    panic!("Can't commit transaction: {:?}", e);
                } else {
                    // TODO: exponential backoff.
                }
            } else {
                break 'read_txn;
            }
        }

        // Drop the database handle.
    }
    DB::destroy(&Options::default(), DB_PATH).unwrap(); // Delete the database directory and files.
}
